import re
import xml.etree.ElementTree as ET
from io import BytesIO
from typing import TextIO, TypedDict


# Define types and interfaces
class ElementModification(TypedDict):

    id: str
    style: dict[str, str]
    text: str


class Replica(TypedDict):

    filename: str
    modifications: list[ElementModification]


Replicas = list[Replica]


# Namespace
ns = {"svg": "http://www.w3.org/2000/svg"}
for prefix, uri in ns.items():
    ET.register_namespace(prefix, uri)


class SvgHandler:

    _tree: ET.ElementTree

    def read(self, file: TextIO) -> None:
        """Reads SVG from TextIO"""
        self._tree = ET.parse(file)

    def modify(self, modifications: list[ElementModification]):
        """Modify elements in SVG by passing list of dicts with shape ElementModification"""
        for modification in modifications:
            if "style" in modification:
                self._modify_element_style(modification["id"], modification["style"])
            if "text" in modification:
                self._set_element_text(modification["id"], modification["text"])

    def get_element_string(self, id: str) -> bytes:
        """Converts element with given id to XML string"""
        # TODO do we want this to be decoded?
        return ET.tostring(self._get_element(id))

    def write(self, file: BytesIO) -> None:
        """Writes SVG to BytesIO"""
        self._tree.write(file)

    def _get_root(self) -> ET.Element:
        try:
            return self._tree.getroot()
        except AttributeError:
            raise Exception("You need to read an SVG file first")

    # All this stuff could be taken to a separate class that wraps ET.Element
    def _get_element(self, id: str) -> ET.Element:
        if (element := self._get_root().find(f".//*[@id='{id}']")) is not None:
            return element
        else:
            raise Exception(f'Element "{id}" not found')

    def _set_element_text(self, id: str, text: str) -> None:
        element = self._get_element(id)
        element.text = text
        # Remove tspan
        if re.match(r"{.*}text", element.tag) and len(element) != 0:
            self._remove_subelements_recursively(element, "svg:tspan")

    def _modify_element_style(self, id: str, style: dict[str, str]) -> None:
        element = self._get_element(id)
        try:
            current_style = self._marshal_style(element.attrib["style"])
        except KeyError:
            current_style = {}
        style = {**current_style, **style}
        element.set("style", self._unmarshal_style(style))

    @staticmethod
    def _remove_subelements_recursively(element: ET.Element, match: str) -> None:
        while (subelement := element.find(match, ns)) is not None:
            element.remove(subelement)

    @staticmethod
    def _marshal_style(style: str) -> dict[str, str]:
        marshaled_style = {}
        for attrib in style.split(";"):
            key, value = tuple(attrib.split(":"))
            marshaled_style[key] = value
        return marshaled_style

    @staticmethod
    def _unmarshal_style(style: dict) -> str:
        return ";".join([f"{key}:{val}" for key, val in style.items()])



import argparse
import json
from typing import Protocol

from .svghandler import Replicas, SvgHandler


class Args(Protocol):

    filename: str
    replications: str


def get_args() -> Args:
    parser = argparse.ArgumentParser(
        description="Replicates and modifies SVG files in batch"
    )
    parser.add_argument(
        "--filename", type=str, help="SVG file that will be used as the basis"
    )
    parser.add_argument(
        "--replicas",
        type=str,
        help="JSON file with specification of replicas that will be created",
    )
    return parser.parse_args()


def main():
    args = get_args()

    with open(args.replicas) as f:
        replicas: Replicas = json.load(f)

    for replica in replicas:
        svg_handler = SvgHandler()

        # Load template
        with open(args.filename) as f:
            svg_handler.read(f)

        # Modify stuff
        svg_handler.modify(replica["modifications"])

        # Save
        with open(replica["filename"], "wb") as f:
            svg_handler.write(f)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['svgreplicate']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['svgreplicate = svgreplicate.__main__:main']}

setup_kwargs = {
    'name': 'svgreplicate',
    'version': '0.1.1',
    'description': 'Tool for batch modifying and replicating SVG-files',
    'long_description': '# SVGReplicate \n\nTool for batch modifying and replicating SVG files. \n\n## Installation \n\nInstall from PyPI by running `pip install svgreplicate`.\n\n## Usage\n\nCreate an SVG file, either by hand or with your favorite vector graphics tool, like [InkScape](https://inkscape.org):\n\n```xml\n<svg\n    width="210mm"\n    height="297mm" \n    id="svg" \n    version="1.1"\n    xmlns="http://www.w3.org/2000/svg" >\n    <g id="group1">\n        <ellipse\n            style="opacity:1;fill:#009bff"\n            id="ellipse1"\n            cx="0.0"\n            cy="0.0"\n            rx="10.0"\n            ry="10.0" />\n        <g id="group2" />\n    <text id="text1">\n        <tspan>First text</tspan>\n        <tspan>Second text \n            <tspan>Nested text</tspan> \n        </tspan>\n    </text>\n    </g>\n</svg>\n```\n\nNext, create a json with the replicas you want: \n\n```json\n[\n    {\n        "filename": "example/replica1.svg", \n        "modifications": [\n        \t{"id": "ellipse1", "style": {"fill": "#000000", "display": "none"}},\n        \t{"id": "group2", "style": {"display": "none"}},\n        \t{"id": "text1", "text": "Hello, world!"}\n        ]\n    },\n    {\n        "filename": "example/replica2.svg",\n        "modifications": [\n                {"id": "ellipse1", "style": {"fill": "#ffffff", "display": "none"}},\n                {"id": "group2", "style": {"display": "none"}},\n                {"id": "text1", "text": "Here\'s Johnny!"}\n        ]\n    }\n]\n```\n\nAnd finally, run the script:\n\n```bash \nsvgreplicate --filename path-to-svg-template.svg --replicas path-to-replicas-config.json \n```\n\nYou now have 2 files in the example folder, based on the template, with the specified modifications. \n\n## Backlog \n\nIn future, I\'d like to add at least the functionality for automatically rendering to PNG. \n\n',
    'author': 'Jan Hein de Jong',
    'author_email': 'janhein.dejong@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/janheindejong/svgreplicate',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
